/***************************************************************************//**
*  \details    This driver will act as an etherbone slave on the host
*              and controls a wishbone master in the FPGA over PCIe
*
*  \author     Michael Reese
*
* *******************************************************************************/
#include<linux/kernel.h>
#include<linux/init.h>
#include<linux/module.h>
#include <linux/kdev_t.h>
#include <linux/err.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/pci.h>


struct pcie_wb_resource {
    unsigned long start;            /* start addr of BAR */
    unsigned long end;          /* end addr of BAR */
    unsigned long size;         /* size of BAR */
    void *addr;             /* remapped addr */
} bar0, bar1;

static int ebs_wbm_pcie_open(struct inode *inode, struct file *file)
{
        printk(KERN_INFO "%s\n", __func__);
        return 0;
}
static int ebs_wbm_pcie_release(struct inode *inode, struct file *file)
{
        printk(KERN_INFO "%s\n", __func__);
        return 0;
}
static ssize_t ebs_wbm_pcie_read(struct file *filp, char __user *buf, size_t len, loff_t *off)
{
    int value;
    printk(KERN_INFO "%s\n", __func__);
    if (len < sizeof(value)) return 0;
    /* write hard coded address into the Direct Access Control Register (DACR) */
    iowrite32(0x4060000, bar0.addr + 0x4); 
    /* make a wishbone read cycle in the hardware */
    value = ioread32(bar1.addr);
    if (copy_to_user(buf, &value, sizeof(value))) {
        return 0;
    }
    return sizeof(value);
}
static ssize_t ebs_wbm_pcie_write(struct file *filp, const char __user *buf, size_t len, loff_t *off)
{
    int value;
    printk(KERN_INFO "%s\n", __func__);
    if (len < sizeof(value)) return 0;
    if (copy_from_user(&value, buf, sizeof(value))) {
        return 0;
    }
    /* write hard coded address into the Direct Access Control Register (DACR) */
    iowrite32(0x4060000, bar0.addr + 0x4); 
    /* make a wishbone write cycle in the hardware */
    iowrite32(value, bar1.addr);
    return sizeof(value);
}

static struct file_operations fops =
{
    .owner      = THIS_MODULE,
    .read       = ebs_wbm_pcie_read,
    .write      = ebs_wbm_pcie_write,
    .open       = ebs_wbm_pcie_open,
    .release    = ebs_wbm_pcie_release,
};

static struct cdev cdev;
dev_t dev = 0;
static struct class *dev_class;
 
static int install_device(void) /* in chapter2 this function was called ebs_wbm_pcie_init */
{
    cdev_init(&cdev,&fops);
    if ((alloc_chrdev_region(&dev, 0, 1, "ebs_wbm_pcie_dev")) <0) {
        return -1;
    }
    if((cdev_add(&cdev,dev,1)) < 0){
        goto r_class;
    }
    dev_class = class_create("ebs_wbm_pcie_class");
    if(IS_ERR(dev_class)){
        goto r_class;
    }
    if(IS_ERR(device_create(dev_class,NULL,dev,NULL,"ebs_wbm_pcie_device"))){
        goto r_device;
    }

    printk(KERN_INFO "Simple kernel module inserted successfully\n");
    return 0;
 
r_device:
    class_destroy(dev_class);
r_class:
    unregister_chrdev_region(dev,1);
    return -1;
}

static void remove_device(void) /* in chapter2 this was called ebs_wbm_pcie_exit */
{
    device_destroy(dev_class,dev);
    class_destroy(dev_class);
    unregister_chrdev_region(dev, 1);
    printk(KERN_INFO "Simple kernel module removed\n");
}


#define PCI_DRIVER_NAME "ebs_wbm" 

static int setup_bar(struct pci_dev* pdev, struct pcie_wb_resource* res, int bar)
{
    res->start = pci_resource_start(pdev, bar);
    res->end   = pci_resource_end(pdev, bar);
    res->size  = res->end - res->start + 1;

    printk(KERN_INFO "setup bar %d  0x%lx - 0x%lx\n", bar, res->start, res->end);

    if ((pci_resource_flags(pdev, 0) & IORESOURCE_MEM) == 0) {
        printk(KERN_INFO "bar %d is not a memory resource\n", bar);
        return -ENOMEM;
    }

    if (!request_mem_region(res->start, res->size, PCI_DRIVER_NAME)) {
        printk(KERN_INFO "bar %d: request_mem_region failed\n", bar);
        return -ENOMEM;
    }

    res->addr = ioremap_cache(res->start, res->size);
    printk(KERN_INFO "bar %d: ioremap to %lx\n", bar, (unsigned long)res->addr);

    return 0;
}

static void destroy_bar(struct pcie_wb_resource* res)
{
    printk(KERN_INFO "released io 0x%lx\n", res->start);

    iounmap(res->addr);
    release_mem_region(res->start, res->size);
}
 
static int probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
    printk(KERN_INFO "probe was called\n");

    if (pci_enable_device(pdev) < 0) {
        printk(KERN_INFO ": could not enable device!\n"); 
        goto r_out;
    }
    if (setup_bar(pdev, &bar0, 0) < 0) {
        printk(KERN_INFO ": could not setup bar0!\n");
        goto r_disable;
    }
    if (setup_bar(pdev, &bar1, 1) < 0) {
        goto r_bar0;
    }
    if (install_device() < 0) {
        goto r_bar1;
    }

    return 0;

r_bar1:
    destroy_bar(&bar1);
r_bar0: 
    destroy_bar(&bar0);
r_disable:
    pci_disable_device(pdev);
r_out:
    return -EIO;
}
static void remove(struct pci_dev *pdev)
{
    remove_device();
    destroy_bar(&bar1);
    destroy_bar(&bar0);
    pci_disable_device(pdev);
    printk(KERN_INFO "remove was called\n");
}

#define PCIE_WB_VENDOR_ID 0x10dc
#define PCIE_WB_DEVICE_ID 0x019a 
static struct pci_device_id ids[] = {
    { PCI_DEVICE(PCIE_WB_VENDOR_ID, PCIE_WB_DEVICE_ID), },
    { 0, }
};
MODULE_DEVICE_TABLE(pci, ids);

static struct pci_driver pcie_wb_driver = {
    .name     = PCI_DRIVER_NAME,
    .id_table = ids,
    .probe    = probe,
    .remove   = remove,
};

static int __init ebs_wbm_pcie_init(void)
{
    return pci_register_driver(&pcie_wb_driver);
}
static void __exit ebs_wbm_pcie_exit(void)
{
    pci_unregister_driver(&pcie_wb_driver);
}

module_init(ebs_wbm_pcie_init);
module_exit(ebs_wbm_pcie_exit);
 
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Michael Reese <m.reese@gsi.de>");
MODULE_DESCRIPTION("Etherbone slave that controls a wishbone master module on the FPGA via PCIe");
MODULE_VERSION("1:0.0");
