/***************************************************************************//**
*  \details    This driver will act as an etherbone slave on the host
*              and controls a wishbone master in the FPGA over PCIe
*
*  \author     Michael Reese
*
* *******************************************************************************/
#include<linux/kernel.h>
#include<linux/init.h>
#include<linux/module.h>
#include <linux/kdev_t.h>
#include <linux/err.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/pci.h>
#include <linux/poll.h>

struct pcie_wb_resource {
    unsigned long start;    /* start addr of BAR */
    unsigned long end;      /* end addr of BAR */
    unsigned long size;     /* size of BAR */
    void *addr;             /* remapped addr */
} bar0, bar1;

void hw_write(uint32_t addr, uint32_t data) {
    iowrite32(addr, bar0.addr + 0x4); /* write addr into DACR */
    iowrite32(data, bar1.addr);       /* write the data */
}
uint32_t hw_read(uint32_t addr) {
    iowrite32(addr, bar0.addr + 0x4); /* write addr into DACR */
    return ioread32(bar1.addr);       /* read the data */
}
uint32_t hw_cfg_read(uint32_t addr) {
    /* this function maps the etherbone slave config register addresses (0x0,0x4,0x8,0xc)
       the the register addresses in the PCIe bridge on the FPGA */
    switch(addr) {
        case 0x0: return ioread32(bar0.addr +  8); /* err shiftreg high    */
        case 0x4: return ioread32(bar0.addr + 12); /* err shiftreg low     */
        case 0x8: return ioread32(bar0.addr + 24); /* wishbone addr offset */
        case 0xc: return ioread32(bar0.addr + 28); /* SDB start address    */
    }
    return 0;
}

#include "ebs.h"
EbSlaveState eb = {
    .hw_write    = hw_write,
    .hw_read     = hw_read,
    .hw_cfg_read = hw_cfg_read
};

static int ebs_wbm_pcie_open(struct inode *inode, struct file *file)
{
    eb_slave_reset(&eb);
    printk(KERN_INFO "%s\n", __func__);
    return 0;
}
static int ebs_wbm_pcie_release(struct inode *inode, struct file *file)
{
    printk(KERN_INFO "%s\n", __func__);
    return 0;
}

#define BUFFERSIZE 8192
char buffer[BUFFERSIZE];
int write_buffer_idx = 0;
int read_buffer_idx = 0;
DECLARE_WAIT_QUEUE_HEAD(read_waitq);
static ssize_t ebs_wbm_pcie_read(struct file *filp, char __user *buf, size_t len, loff_t *off)
{
    printk(KERN_INFO "%s\n", __func__);
    if (read_buffer_idx == write_buffer_idx) { /* read was called but no data is available */
        if (filp->f_flags & O_NONBLOCK) { /* device is nonblocking -> return -EAGAIN */
            return -EAGAIN;
        } 
        /* device is blocking -> wait_event_interruptible */
        wait_event_interruptible(read_waitq, (read_buffer_idx<write_buffer_idx));
    }
    if (len > (write_buffer_idx-read_buffer_idx)) {
        len = write_buffer_idx-read_buffer_idx;
    }
    if (copy_to_user(buf, buffer+read_buffer_idx, len)) {
        return 0;
    }
    read_buffer_idx += len;
    if (read_buffer_idx == write_buffer_idx) {
        read_buffer_idx  = 0;
        write_buffer_idx = 0;
    }
    return len;
}
static ssize_t ebs_wbm_pcie_write(struct file *filp, const char __user *buf, size_t len, loff_t *off)
{
    printk(KERN_INFO "%s\n", __func__);
    if (len > BUFFERSIZE) len = BUFFERSIZE;
    if (copy_from_user(buffer, buf, len)) {
        return 0;
    }
    write_buffer_idx = len;
    for (int i = 0; i < write_buffer_idx; i += 4) {
        uint32_t request, response;
        request = *(int*)&buffer[i];
        eb_slave_process(&eb, &buffer[i]);
        response = *(int*)&buffer[i];
        printk(KERN_INFO "0x%08x -> 0x%08x\n", be32_to_cpu(request), be32_to_cpu(response));
    }
    wake_up_interruptible(&read_waitq);
    return len;
}
static unsigned int ebs_wbm_pcie_poll(struct file *filp, struct poll_table_struct *wait)
{
    unsigned int mask = 0;
    printk(KERN_INFO "%s\n", __func__);
    poll_wait(filp, &read_waitq, wait);
    if (read_buffer_idx < write_buffer_idx)  {
        mask |= POLLIN | POLLRDNORM; /* readable if read cursor is left of write cursor */
    }
    if (write_buffer_idx == 0) {
        mask |= POLLOUT | POLLWRNORM; /* always writable */
    }

    return mask;
}

static struct file_operations fops =
{
    .owner      = THIS_MODULE,
    .read       = ebs_wbm_pcie_read,
    .write      = ebs_wbm_pcie_write,
    .poll       = ebs_wbm_pcie_poll,
    .open       = ebs_wbm_pcie_open,
    .release    = ebs_wbm_pcie_release,
};

static struct cdev cdev;
dev_t dev = 0;
static struct class *dev_class;
 
static int install_device(void) 
{
    cdev_init(&cdev,&fops);
    if ((alloc_chrdev_region(&dev, 0, 1, "ebs_wbm_pcie_dev")) <0) {
        return -1;
    }
    if((cdev_add(&cdev,dev,1)) < 0){
        goto r_class;
    }
    dev_class = class_create("ebs_wbm_pcie_class");
    if(IS_ERR(dev_class)){
        goto r_class;
    }
    if(IS_ERR(device_create(dev_class,NULL,dev,NULL,"ebs_wbm_pcie_device"))){
        goto r_device;
    }

    printk(KERN_INFO "Simple kernel module inserted successfully\n");
    return 0;
 
r_device:
    class_destroy(dev_class);
r_class:
    unregister_chrdev_region(dev,1);
    return -1;
}

static void remove_device(void) 
{
    device_destroy(dev_class,dev);
    class_destroy(dev_class);
    unregister_chrdev_region(dev, 1);
    printk(KERN_INFO "Simple kernel module removed\n");
}


#define PCI_DRIVER_NAME "ebs_wbm" 

static int setup_bar(struct pci_dev* pdev, struct pcie_wb_resource* res, int bar)
{
    res->start = pci_resource_start(pdev, bar);
    res->end   = pci_resource_end(pdev, bar);
    res->size  = res->end - res->start + 1;

    printk(KERN_INFO "setup bar %d  0x%lx - 0x%lx\n", bar, res->start, res->end);

    if ((pci_resource_flags(pdev, 0) & IORESOURCE_MEM) == 0) {
        printk(KERN_INFO "bar %d is not a memory resource\n", bar);
        return -ENOMEM;
    }

    if (!request_mem_region(res->start, res->size, PCI_DRIVER_NAME)) {
        printk(KERN_INFO "bar %d: request_mem_region failed\n", bar);
        return -ENOMEM;
    }

    res->addr = ioremap_cache(res->start, res->size);
    printk(KERN_INFO "bar %d: ioremap to %lx\n", bar, (unsigned long)res->addr);

    return 0;
}

static void destroy_bar(struct pcie_wb_resource* res)
{
    printk(KERN_INFO "released io 0x%lx\n", res->start);

    iounmap(res->addr);
    release_mem_region(res->start, res->size);
}
static void wb_set_interrupt(int on)
{
    static int CONTROL_REGISTER_HIGH  = 0;
    uint32_t value  = 0x10000000;
    if (on)  value |= 0x20000000;
    iowrite32(value, bar0.addr + CONTROL_REGISTER_HIGH);
}
static int wb_read_msi_data(void)
{
    static int MASTER_CTL_HIGH = 64; 
    static int MASTER_ADR_LOW  = 76;
    static int MASTER_DAT_LOW  = 84;
    uint32_t ctl       = ioread32(bar0.addr + MASTER_CTL_HIGH);
    uint32_t msi_addr  = ioread32(bar0.addr + MASTER_ADR_LOW);
    uint32_t msi_data  = ioread32(bar0.addr + MASTER_DAT_LOW);
    uint32_t msi_valid = (ctl & 0x80000000) != 0;
    if (msi_valid) {
        printk(KERN_INFO "got IRQ data: %08x %08x msi_valid=%d\n", msi_addr, msi_data, msi_valid);
        iowrite32(1, bar0.addr + MASTER_CTL_HIGH); /* pop the data from the hardware queue */
    }
    return msi_valid;
}
static irqreturn_t irq_handler(int irq, void *dev_id)
{
    printk(KERN_INFO "%s\n", __func__);
    wb_set_interrupt(0); /* disable IRQ */
    wb_read_msi_data();
    wb_set_interrupt(1); /* enable IRQ */
    return IRQ_HANDLED;
} 
static int probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
    printk(KERN_INFO "probe was called\n");

    if (pci_enable_device(pdev) < 0) {
        printk(KERN_INFO ": could not enable device!\n"); 
        goto r_out;
    }
    if (setup_bar(pdev, &bar0, 0) < 0) {
        printk(KERN_INFO ": could not setup bar0!\n");
        goto r_device;
    }
    if (setup_bar(pdev, &bar1, 1) < 0) {
        goto r_bar0;
    }

    pci_set_master(pdev); 
    if (pci_enable_msi(pdev) < 0) {
        goto r_bar1;
    }
    if (request_irq(pdev->irq, irq_handler, IRQF_SHARED, "ebs_wbm_pcie", pdev) < 0) {
        goto r_msi;
    }
    while (wb_read_msi_data()); /* empty the queue */
    wb_set_interrupt(1);

    if (install_device() < 0) {
        goto r_irq;
    }

    return 0;

r_irq:
    free_irq(pdev->irq, pdev);
r_msi:
    pci_disable_msi(pdev);
r_bar1:
    pci_clear_master(pdev);
    destroy_bar(&bar1);
r_bar0: 
    destroy_bar(&bar0);
r_device:
    pci_disable_device(pdev);
r_out:
    return -EIO;
}

static void remove(struct pci_dev *pdev)
{
    remove_device();
    free_irq(pdev->irq, pdev);
    pci_disable_msi(pdev);
    pci_clear_master(pdev);
    destroy_bar(&bar1);
    destroy_bar(&bar0);
    pci_disable_device(pdev);
    printk(KERN_INFO "remove was called\n");
}

#define PCIE_WB_VENDOR_ID 0x10dc
#define PCIE_WB_DEVICE_ID 0x019a 
static struct pci_device_id ids[] = {
    { PCI_DEVICE(PCIE_WB_VENDOR_ID, PCIE_WB_DEVICE_ID), },
    { 0, }
};
MODULE_DEVICE_TABLE(pci, ids);

static struct pci_driver pcie_wb_driver = {
    .name     = PCI_DRIVER_NAME,
    .id_table = ids,
    .probe    = probe,
    .remove   = remove,
};

static int __init ebs_wbm_pcie_init(void)
{
    printk(KERN_INFO "%s\n", __func__);
    return pci_register_driver(&pcie_wb_driver);
}
static void __exit ebs_wbm_pcie_exit(void)
{
    printk(KERN_INFO "%s\n", __func__);
    pci_unregister_driver(&pcie_wb_driver);
}

module_init(ebs_wbm_pcie_init);
module_exit(ebs_wbm_pcie_exit);
 
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Michael Reese <m.reese@gsi.de>");
MODULE_DESCRIPTION("Etherbone slave that controls a wishbone master module on the FPGA via PCIe");
MODULE_VERSION("1:0.0");
